/**
 * EasyBeans
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ActivatorWithBackedClassLoader.java 1907 2007-10-10 14:36:48Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.osgi.ejbjar;

import java.net.URL;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.ow2.easybeans.api.EZBContainer;
import org.ow2.easybeans.api.EZBContainerConfig;
import org.ow2.easybeans.api.EZBContainerException;
import org.ow2.easybeans.api.EZBServer;
import org.ow2.easybeans.container.JContainer3;
import org.ow2.easybeans.container.JContainerConfig;
import org.ow2.easybeans.osgi.util.BCMapper;
import org.ow2.easybeans.osgi.util.LDAPFilter;
import org.ow2.easybeans.server.Embedded;
import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This should be a Generic EJB3.0 ActivatorWithBackedClassLoader.
 */
public class Activator implements BundleActivator {

    /**
     * Root of the bundle.
     */
    private static final String ROOT = "/";

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(Activator.class);

    /**
     * The local JContainer instance (ie the EjbJar).
     */
    private EZBContainer container = null;

    /**
     * BundleContext.
     */
    private BundleContext bc = null;

    /**
     * The Embedded listener.
     */
    private ServiceListener listener;

    /**
     * ServiceRegistration object.
     */
    private ServiceRegistration serviceReg = null;

    /**
     * Register ServiceListener and start the COntainer if an Embedded instance is available. {@inheritDoc}
     *
     * @see org.osgi.framework.BundleActivator#start(org.osgi.framework.BundleContext)
     */
    public void start(final BundleContext bc) throws Exception {
        this.bc = bc;
        this.listener = new EmbeddedServiceListener(this);

        String filter = LDAPFilter.createLDAPFilter(Embedded.class.getName());
        bc.addServiceListener(listener, filter);

        // first init, try to get the Embedded reference
        ServiceReference sr = bc.getServiceReference(Embedded.class.getName());
        if (sr != null) {
            // and use it only if the service is available
            startContainer(sr);
        }
    }

    /**
     * Unregister ServiceListener and stop the Container. {@inheritDoc}
     *
     * @see org.osgi.framework.BundleActivator#stop(org.osgi.framework.BundleContext)
     */
    public void stop(final BundleContext bc) throws Exception {
        bc.removeServiceListener(listener);
        // Stop the container if any
        stopContainer();
    }

    /**
     * Start the EZBContainer and register the instance as an OSGi Service.
     *
     * @param sr
     *            The Embedded ServiceReference.
     */
    public void startContainer(final ServiceReference sr) {
        if (container == null) {
            ClassLoader old = Thread.currentThread().getContextClassLoader();

            try {
                // Create the EZBArchive relying on OSGi
                ArchiveManager am = ArchiveManager.getInstance();
                IArchive archive = am.getArchive(bc);

                BCMapper.getInstance().put(bc.getBundle().getEntry(ROOT), bc);

                BackedClassLoader backedClassLoader = new BackedClassLoader(new URL[] {archive.getURL()}, this.getClass()
                        .getClassLoader(), bc);

                Thread.currentThread().setContextClassLoader(backedClassLoader);

                logger.info("Creating Container from the Bundle Archive ''{0}''", archive.getURL());

                // Create and start the container
                EZBContainerConfig jConfig = new JContainerConfig(archive);
                jConfig.setEZBServer((EZBServer) bc.getService(sr));
                container = new JContainer3(jConfig);

                container.addExtension(BundleContext.class, bc);

                container.setClassLoader(backedClassLoader);

                container.start();

                // register the container as a Service
                serviceReg = bc.registerService(EZBContainer.class.getName(), container, null);

            } catch (EZBContainerException e) {
                if (container != null) {
                    container.stop();
                }
                BCMapper.getInstance().remove(bc.getBundle().getEntry(ROOT));
                throw new RuntimeException(e);
            } catch (ArchiveException e) {
                if (container != null) {
                    container.stop();
                }
                BCMapper.getInstance().remove(bc.getBundle().getEntry(ROOT));
                throw new RuntimeException(e);
            } finally {
                Thread.currentThread().setContextClassLoader(old);
            }
        }
    }

    /**
     * Stop the EZBContainer and unregister the exposed Service.
     */
    public void stopContainer() {
        // Remove the container
        if (container != null) {

            ClassLoader old = Thread.currentThread().getContextClassLoader();

            try {
                Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());

                // Stop
                container.stop();

                container.removeExtension(BundleContext.class);

                // unregister if started
                if (serviceReg != null) {
                    serviceReg.unregister();
                }

                BCMapper.getInstance().remove(bc.getBundle().getEntry(ROOT));

                container = null;

            } finally {
                Thread.currentThread().setContextClassLoader(old);
            }

        }
    }

}
